#region License

/* Copyright (c) 2000-2008 Marc Jacobi
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */

#endregion

using System;
using System.Diagnostics;
using System.Reflection;
using System.Text;

namespace Jacobi.Diagnostics.Formatters
{
    /// <summary>
    /// Formats a Trace text for a <see cref="MethodBase"/>.
    /// </summary>
    internal class MethodBaseTraceFormatter : TraceFormatterBase
    {
        public override string Format(TraceContext context, object objectToFormat)
        {
            return Format(context, (MethodBase)objectToFormat);
        }

        public string Format(TraceContext context, MethodBase methodBase)
        {
            StringBuilder buffer = new StringBuilder();

            FormatVisibility(buffer, methodBase);
            FormatReturnAndName(buffer, methodBase);
            FormatParameters(context, buffer, methodBase);
            FormatReturnValue(context, buffer, methodBase);

            return buffer.ToString();
        }

        private void FormatReturnValue(TraceContext context, StringBuilder buffer, MethodBase methodBase)
        {
            if (context != null && context.ReturnValue != null)
            {
                buffer.Append("='");
                buffer.Append(context.ReturnValue);
                buffer.Append("'");
            }
        }

        private void FormatParameters(TraceContext context, StringBuilder buffer, MethodBase methodBase)
        {
            ParameterInfo[] paramInfos = null;
            MethodInfo methodInfo = methodBase as MethodInfo;

            if (methodInfo != null)
            {
                paramInfos = methodInfo.GetParameters();
            }
            else
            {
                ConstructorInfo ctorInfo = methodBase as ConstructorInfo;

                if (ctorInfo != null)
                {
                    paramInfos = ctorInfo.GetParameters();
                }
            }

            buffer.Append("(");

            if (paramInfos != null && paramInfos.Length > 0)
            {
                int index = 0;
                foreach (ParameterInfo paramInfo in paramInfos)
                {
                    if (index > 0)
                    {
                        buffer.Append(", ");
                    }

                    FormatParameter(context, buffer, paramInfo);

                    index++;
                }
            }

            buffer.Append(")");
        }

        private void FormatParameter(TraceContext context, StringBuilder buffer, ParameterInfo paramInfo)
        {
            if (paramInfo.IsIn && paramInfo.IsOut)
            {
                buffer.Append("ref ");
            }
            else if (paramInfo.IsOut)
            {
                buffer.Append("out ");
            }

            buffer.Append(paramInfo.ParameterType.Name);
            buffer.Append(" ");
            buffer.Append(paramInfo.Name);

            if (context != null)
            {
                if (context.Parameters.Contains(paramInfo.Name))
                {
                    MethodContext.ParameterInfo paramValueInfo =
                        context.Parameters[paramInfo.Name];

                    buffer.Append("=");
                    if (paramValueInfo.Value == null)
                    {
                        buffer.Append("null");
                    }
                    else
                    {
                        buffer.Append("'");
                        buffer.Append(paramValueInfo.Value);
                        buffer.Append("'");
                    }
                }
            }
        }

        private void FormatReturnAndName(StringBuilder buffer, MethodBase methodBase)
        {
            MethodInfo methodInfo = methodBase as MethodInfo;

            if (methodInfo != null)
            {
                buffer.Append(methodInfo.ReturnType.Name);
                buffer.Append(" ");
            }

            buffer.Append(methodBase.Name);
        }

        private void FormatVisibility(StringBuilder buffer, MethodBase methodBase)
        {
            if (methodBase.IsPublic)
            {
                buffer.Append("public ");
            }
            else if (methodBase.IsPrivate)
            {
                buffer.Append("private ");
            }
            else if(methodBase.IsFamilyOrAssembly)
            {
                buffer.Append("internal protected ");
            }
            else if (methodBase.IsAssembly)
            {
                buffer.Append("internal ");
            }
            else if(methodBase.IsFamily)
            {
                buffer.Append("protected ");
            }

            if (methodBase.IsAbstract)
            {
                buffer.Append("abstract ");
            }
            else if (methodBase.IsStatic)
            {
                buffer.Append("static ");
            }
            else if (methodBase.IsVirtual)
            {
                buffer.Append("virtual ");
            }

            // TODO: new and override

            if (methodBase.IsFinal)
            {
                buffer.Append("sealed ");
            }
        }
    }
}
