#region License

/* Copyright (c) 2000-2008 Marc Jacobi
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */

#endregion

using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Reflection;
using System.Runtime.CompilerServices;
using System.Runtime.Remoting.Messaging;

using Jacobi;

namespace Jacobi.Diagnostics
{
    /// <summary>
    /// The MethodContext class gathers context information during
    /// execution of a method.
    /// </summary>
    /// <remarks>Context does not automagically gather information behind the scenes. 
    /// It has to be added explicitly using the public methods.
    /// All references to other objects (parameters, return value and local vars)
    /// are managed as <see cref="WeakReference"/>s.</remarks>
    public class MethodContext : DisposableBase
    {
        private const string MethodContextSlotName = "CurrentMethodContext";
        private const string ReturnValueDataKey = "_return";
        private const string ClassInstanceDataKey = "_this";

        /// <summary>
        /// Constructs a new instance.
        /// </summary>
        /// <param name="skipFrames">The number of stack frames to skip
        /// in order to determine the calling (or current) method.</param>
        /// <remarks>If a new MethodContext instance is created directly
        /// in the method pass 0 for <paramref name="skipFrames"/>. 
        /// If a custom class derives from MethodContext pass 1 to bypass
        /// the custom class' constructor.</remarks>
        [MethodImpl(MethodImplOptions.NoInlining)]
        public MethodContext(int skipFrames)
        {
            Throw.IfArgumentOutOfRange(skipFrames, 0, Int32.MaxValue, "skipFrames");

            _currentFrame = new StackFrame(skipFrames + 1, true);
            PushMethodContext();
        }

        /// <summary>
        /// Constructs an instance on a known method.
        /// </summary>
        /// <param name="methodBase">Must not be null.</param>
        protected MethodContext(MethodBase methodBase)
        {
            Throw.IfArgumentNull(methodBase, "methodBase");

            _methodBase = methodBase;

            // NOTE: this constructor leaves _currentFrame null!
            PushMethodContext();
        }

        private StackFrame _currentFrame;
        private MethodBase _methodBase;
        /// <summary>
        /// Gets the Method information of the current method.
        /// </summary>
        /// <remarks>If this property does not return the expected
        /// method information, you've passed a wrong <b>skipFrames</b>
        /// parameter to the constructor.</remarks>
        public MethodBase MethodBase
        {
            get
            {
                ThrowIfDisposed();

                if (_methodBase == null && _currentFrame != null)
                {
                    _methodBase = _currentFrame.GetMethod();
                }

                return _methodBase;
            }
        }

        private object _instance;
        /// <summary>
        /// Gets or sets the instance the method is called on.
        /// </summary>
        /// <remarks>Is null for static methods.</remarks>
        public object ClassInstance
        {
            get { return _instance; }
            protected set { _instance = value; }
        }

        private ParameterInfoCollection _parameters;
        /// <summary>
        /// Gets a collection method parameters.
        /// </summary>
        /// <remarks>Use this collection or the <see cref="M:AddParameterValue"/>
        /// method to add information on the method's parameter values.</remarks>
        public ParameterInfoCollection Parameters
        {
            get
            {
                ThrowIfDisposed();

                if (_parameters == null)
                {
                    _parameters = new ParameterInfoCollection();
                }

                return _parameters;
            }
        }

        /// <summary>
        /// Adds a method parameter name and its value to the 
        /// <see cref="Parameters"/> collection.
        /// </summary>
        /// <typeparam name="T">The Type of the parameter value (inferred).</typeparam>
        /// <param name="value">The value of the parameter.</param>
        /// <param name="paramName">The name of the parameter. Must be unique.</param>
        public void AddParameterValue<T>(T value, string paramName)
        {
            ThrowIfDisposed();
            Throw.IfArgumentNullOrEmpty(paramName, "paramName");

            ParameterInfo paramInfo = new ParameterInfo(paramName, value);
            Parameters.Add(paramInfo);
        }

        private WeakReference _returnValue;
        /// <summary>
        /// Gets or sets the return value of the method.
        /// </summary>
        /// <remarks>Note that the value can be null when
        /// the retrun type of the method is Void or null was actually
        /// returned from the method.</remarks>
        public object ReturnValue
        {
            get
            {
                ThrowIfDisposed();

                if (_returnValue == null)
                {
                    return null;
                }

                return _returnValue.Target;
            }
            set
            {
                ThrowIfDisposed();

                if (value != null)
                {
                    _returnValue = new WeakReference(value);
                }
                else
                {
                    _returnValue = null;
                }
            }
        }

        /// <summary>
        /// Sets the <see cref="P:ReturnValue"/>.
        /// </summary>
        /// <typeparam name="T">The type of the return value (inferred).</typeparam>
        /// <param name="value">The return value. Can be null.</param>
        public void SetReturnValue<T>(T value)
        {
            ThrowIfDisposed();

            ReturnValue = value;
        }

        private ParameterInfoCollection _localVars;
        /// <summary>
        /// Gets the collection of local variables of the method.
        /// </summary>
        /// /// <remarks>Use this collection or the <see cref="M:AddLocalVariable"/>
        /// method to add information on the method's local var values.</remarks>
        public ParameterInfoCollection LocalVariables
        {
            get
            {
                ThrowIfDisposed();

                if (_localVars == null)
                {
                    _localVars = new ParameterInfoCollection();
                }

                return _localVars;
            }
        }

        /// <summary>
        /// Adds a method's local var name and its value to the 
        /// <see cref="LocalVariables"/> collection.
        /// </summary>
        /// <typeparam name="T">The Type of the local var value (inferred).</typeparam>
        /// <param name="value">The value of the local var.</param>
        /// <param name="varName">The name of the local var. Must be unique.</param>
        public void AddLocalVariable<T>(T value, string varName)
        {
            ThrowIfDisposed();
            Throw.IfArgumentNullOrEmpty(varName, "varName");

            ParameterInfo paramInfo = new ParameterInfo(varName, value);
            LocalVariables.Add(paramInfo);
        }

        private WeakReference _lastError;
        /// <summary>
        /// Gets or sets the last error that was caught in a method.
        /// </summary>
        public Exception LastError
        {
            get
            {
                ThrowIfDisposed();

                if (_lastError == null)
                {
                    return null;
                }

                return (Exception)_lastError.Target;
            }
            set
            {
                ThrowIfDisposed();

                if (value != null)
                {
                    _lastError = new WeakReference(value);
                }
                else
                {
                    _lastError = null;
                }
            }
        }

        private MethodContext _callingMethodCtx;
        /// <summary>
        /// Gets the <see cref="MethodContext"/> of the calling method.
        /// </summary>
        /// <remarks>
        /// The stack of calling methods can be different than the 
        /// actual call stack. If not all methods in the call stack use
        /// a MethodContext they will be skipped in this (linked) list.
        /// </remarks>
        public MethodContext CallingContext
        {
            get { return _callingMethodCtx; }
        }

        /// <summary>
        /// Transfers the parameter and return value information
        /// to the <see cref="P:System.Exception.Data"/> dictionary.
        /// </summary>
        /// <param name="e">The exception to add context to. Can be null.</param>
        /// <remarks>The <see cref="P:ReturnValue"/> is filed under "_return" and will 
        /// overwrite that slot if it is already present. The <see cref="ClassInstance"/>
        /// is filed under "_this" and will also overwrite an existing slot.
        /// The instances are only added as strong (normal) references when the type is serializable.
        /// </remarks>
        public void AddContextTo(Exception e)
        {
            ThrowIfDisposed();

            if (e != null)
            {
                if (_instance != null && _instance.GetType().IsSerializable)
                {
                    e.Data[ClassInstanceDataKey] = _instance;
                }

                if (_parameters != null)
                {
                    foreach (ParameterInfo paramInfo in Parameters)
                    {
                        if (paramInfo.Value.GetType().IsSerializable)
                        {
                            e.Data[paramInfo.Name] = paramInfo.Value;
                        }
                    }
                }

                if (_localVars != null)
                {
                    foreach (ParameterInfo varInfo in LocalVariables)
                    {
                        if (varInfo.Value.GetType().IsSerializable)
                        {
                            e.Data[varInfo.Name] = varInfo.Value;
                        }
                    }
                }

                if (_returnValue != null && _returnValue.GetType().IsSerializable)
                {
                    e.Data[ReturnValueDataKey] = ReturnValue;
                }
            }
        }

        // pushes this MethodContext onto the stack
        private void PushMethodContext()
        {
            Debug.Assert(_callingMethodCtx == null);

            _callingMethodCtx = (MethodContext)
                CallContext.LogicalGetData(MethodContextSlotName);

            CallContext.LogicalSetData(MethodContextSlotName, this);
        }

        // pops the latest MethodContext from the stack and throws an exception if it is not this instance.
        private void PopMethodContext()
        {
            MethodContext thisCtx = (MethodContext)
                CallContext.LogicalGetData(MethodContextSlotName);

            if (thisCtx != this)
            {
                throw new InvalidOperationException(
                    "The MethodContext stack is corrupted!");
            }

            CallContext.LogicalSetData(MethodContextSlotName, _callingMethodCtx);
            
            _callingMethodCtx = null;
        }

        protected static int SkipAdditionalFramesForCurrentContext = 0;
        /// <summary>
        /// Returns the current MethodContext or null.
        /// </summary>
        /// <param name="skipFrames">The number of call stack frames to skip
        /// to determine if the MethodContext on the stack belongs to
        /// the calling method.</param>
        /// <returns>Returns null if the MethodContext on the stack does
        /// not belong to the calling method or the stack is empty.</returns>
        [MethodImpl(MethodImplOptions.NoInlining)]
        protected internal static MethodContext GetCurrent(int skipFrames)
        {
            Throw.IfArgumentOutOfRange(skipFrames, 0, Int32.MaxValue, "skipFrames");

            MethodContext ctx = (MethodContext)
                CallContext.LogicalGetData(MethodContextSlotName);

            if (ctx != null)
            {
                StackFrame callingFrame = new StackFrame(skipFrames + 1 + SkipAdditionalFramesForCurrentContext, true);
                if (ctx.MethodBase != callingFrame.GetMethod())
                {
                    // current context is not for the calling method.
                    ctx = null;
                }
            }

            return ctx;
        }

        /// <summary>
        /// Override by derived class to perform cleanup and context shutdown.
        /// </summary>
        /// <param name="disposing">Always true.</param>
        protected override void Dispose(bool disposing)
        {
            try
            {
                if (disposing)
                {
                    // we assume that if Dispose is called from a finalizer, 
                    // it is no use trying to sync the call stack.
                    PopMethodContext();
                }
            }
            finally
            {
                base.Dispose(disposing);
            }
        }

        //---------------------------------------------------------------------

        /// <summary>
        /// Manages a <see cref="WeakReference"/> to a named value.
        /// </summary>
        /// <remarks>The struct is immutable after construction.</remarks>
        public struct ParameterInfo
        {
            /// <summary>
            /// Constructs a new initialized instance.
            /// </summary>
            /// <param name="name">The name of the parameter.</param>
            /// <param name="value">The value of the parameter.</param>
            public ParameterInfo(string name, object value)
            {
                _name = name;
                _value = new WeakReference(value);
            }

            private string _name;
            /// <summary>
            /// Gets the parameter name.
            /// </summary>
            public string Name
            {
                get { return _name; }
            }

            private WeakReference _value;
            /// <summary>
            /// Gets the parameter value.
            /// </summary>
            /// <remarks>Can be null.</remarks>
            public object Value
            {
                get { return (_value == null) ? null : _value.Target; }
            }
        }

        //---------------------------------------------------------------------

        /// <summary>
        /// Manages a collection of <see cref="ParameterInfo"/> instances.
        /// </summary>
        public class ParameterInfoCollection : KeyedCollection<string, ParameterInfo>
        {
            /// <summary>
            /// Returns the unique identifying name for a parameter info struct.
            /// </summary>
            /// <param name="item"></param>
            /// <returns></returns>
            protected override string GetKeyForItem(ParameterInfo item)
            {
                return item.Name;
            }
        }
    }
}
