#region License

/* Copyright (c) 2000-2008 Marc Jacobi
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */

#endregion

using System;
using System.Collections.Generic;
using System.Diagnostics;

namespace Jacobi.Diagnostics
{
    /// <summary>
    /// The MethodTraceSwitch class is a <see cref="TraceSwitch"/> that
    /// adds a <see cref="MethodTraceFlow"/> filter.
    /// </summary>
    /// <remarks>This class is used by default to construct new method
    /// TraceSwitches. Through the use of a <see cref="SwitchAttribute"/>
    /// on the method, the type of switch instantiated can be cusomized
    /// (as well as its name). The specified type should derive from
    /// <see cref="TraceSwitch"/>, though.
    /// In config use the <b>traceFlow</b> to specify the comma separated
    /// <see cref="MethodTraceFlow"/> values.
    /// </remarks>
    public class MethodTraceSwitch : TraceSwitch
    {
        private const string TraceFlowAttributeName = "traceFlow";

        /// <summary>
        /// Constructs a new initialized instance.
        /// </summary>
        /// <param name="displayName">The (config) name of the method switch.</param>
        /// <param name="description">A description for the switch.</param>
        public MethodTraceSwitch(string displayName, string description)
            : base(displayName, description)
        { }

        /// <summary>
        /// Constructs a new initialized instance.
        /// </summary>
        /// <param name="displayName">The (config) name of the method switch.</param>
        /// <param name="description">A description for the switch.</param>
        /// <param name="defaultSwitchLevel">The default <see cref="TraceLevel"/>.</param>
        public MethodTraceSwitch(string displayName, string description, string defaultSwitchLevel)
            : base(displayName, description, defaultSwitchLevel)
        { }

        /// <summary>
        /// Returns a string array of custom config attributes.
        /// </summary>
        /// <returns>Never returns null.</returns>
        protected override string[] GetSupportedAttributes()
        {
            List<string> attrs = new List<string>();

            string[] baseAttrs = base.GetSupportedAttributes();
            if(baseAttrs != null)
            {
                attrs.AddRange(baseAttrs);
            }

            attrs.Add(TraceFlowAttributeName);

            return attrs.ToArray();
        }

        /// <summary>
        /// Handles the state change when the (config) settings
        /// have changed.
        /// </summary>
        protected override void OnSwitchSettingChanged()
        {
            base.OnSwitchSettingChanged();

            if (base.Attributes.ContainsKey(TraceFlowAttributeName))
            {
                FlowLevel = (MethodTraceFlow)
                    Enum.Parse(typeof(MethodTraceFlow),
                    base.Attributes[TraceFlowAttributeName]);
            }
            else
            {
                FlowLevel = MethodTraceFlow.Entry | MethodTraceFlow.Exit;
            }
        }

        private MethodTraceFlow _flowLevel;
        /// <summary>
        /// Gets or sets the <b>traceFlow</b> value.
        /// </summary>
        public MethodTraceFlow FlowLevel
        {
            get { return _flowLevel; }
            set { _flowLevel = value; }
        }

        /// <summary>
        /// Returns true when the method entry should be traced.
        /// </summary>
        public bool TraceMethodEntry
        {
            get { return ((FlowLevel & MethodTraceFlow.Entry) > 0); }
        }

        /// <summary>
        /// Returns true when the method exit should be traced.
        /// </summary>
        public bool TraceMethodExit
        {
            get { return ((FlowLevel & MethodTraceFlow.Exit) > 0); }
        }
    }
}
