#region License

/* Copyright (c) 2000-2008 Marc Jacobi
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */

#endregion

using System;
using System.Diagnostics;
using System.Runtime.CompilerServices;

namespace Jacobi.Diagnostics
{
    /// <summary>
    /// The MethodTracer class is responsible for outputting traces
    /// for one specific method.
    /// </summary>
    /// <remarks>The MethodTracer class combines the <see cref="TraceSource"/>
    /// for the <see cref="System.Type"/> the method belongs to with a
    /// <see cref="TraceSwitch"/> to further allow trace filtering on
    /// method level.</remarks>
    public class MethodTracer
    {
        /// <summary>
        /// Constructs a new instance base on a TraceSource and a TraceSwitch
        /// </summary>
        /// <param name="traceSource">Must not be null.</param>
        /// <param name="traceSwitch">Can be null.</param>
        internal MethodTracer(TraceSource traceSource, TraceSwitch traceSwitch)
        {
            Throw.IfArgumentNull(traceSource, "traceSource");
            //Throw.IfArgumentNull(traceSwitch, "traceSwitch");

            _traceSource = traceSource;
            _traceSwitch = traceSwitch;
        }

        /// <summary>
        /// Determines if the <paramref name="traceFlow"/> is enabled by the TraceSwitch;
        /// </summary>
        /// <param name="traceFlow">A method flow tracing level.</param>
        /// <returns>Returns true if the specified level should be output to the TraceSource.</returns>
        internal bool IsTraceEnabled(MethodTraceFlow traceFlow)
        {
            bool enabled = true;

            MethodTraceSwitch methodSwitch = TraceSwitch as MethodTraceSwitch;

            if (methodSwitch != null)
            {
                enabled = ((methodSwitch.FlowLevel & traceFlow) > 0);
            }

            return enabled;
        }

        private TraceSwitch _traceSwitch;
        /// <summary>
        /// Gets the TraceSwitch that indicates the level of tracing enabled.
        /// </summary>
        internal TraceSwitch TraceSwitch
        {
            get { return _traceSwitch; }
        }

        private TraceSource _traceSource;
        /// <summary>
        /// Gets the TraceSource that represents a channel to write trace text to.
        /// </summary>
        internal TraceSource TraceSource
        {
            get { return _traceSource; }
        }

        /// <summary>
        /// Outputs a formatted trace text using the specified <paramref name="context"/>.
        /// </summary>
        /// <param name="context">The context contains information used in formatting text.</param>
        /// <param name="traceEventType">The type of event this trace represents.</param>
        /// <param name="format">An optional format specifyer that contains '{0}'. Can be null.</param>
        /// <param name="state">A state object to format. Can be null.</param>
        /// <param name="id">An event id. Is misused to trace the instance hash code sometimes.</param>
        //[Conditional("TRACE")]
        internal void FormatTrace(TraceContext context, TraceEventType traceEventType, string format, object state, int id)
        {
            if (!IsEnabled(traceEventType)) return;

            // calls trace formatters to format the state object
            string traceText = TraceManager.Current.Format(context, state);

            if (!String.IsNullOrEmpty(format))
            {
                traceText = String.Format(format, traceText);
            }

            TraceSource.TraceEvent(traceEventType, id, traceText);
        }

        /// <summary>
        /// Formats a trace text to output.
        /// </summary>
        /// <param name="traceEventType">The type of trace event.</param>
        /// <param name="format">A format specifyer containing '{0}'. Can be null.</param>
        /// <param name="state">A state object to format. Can be null.</param>
        /// <param name="id">An event id. Is misused to trace the instance hash code sometimes.</param>
        /// <remarks>
        /// The current <see cref="TraceContext"/> is fetched to get a MethodTracer instance
        /// for the calling method. If no valid context could be found the <see cref="TraceManager.DefaultMethodTracer"/>
        /// is used and no method context information will be availabe.
        /// This method is only active when the 'TRACE' directive is declared.
        /// </remarks>
        [Conditional("TRACE")]
        [MethodImpl(MethodImplOptions.NoInlining)]
        public static void FormatTrace(TraceEventType traceEventType, string format, object state, int id)
        {
            TraceContext ctx = TraceContext.GetCurrent(1);
            MethodTracer tracer = null;

            if (ctx != null)
            {
                tracer = ctx.MethodTracer;
            }
            else
            {
                tracer = TraceManager.Current.DefaultMethodTracer;
            }

            tracer.FormatTrace(ctx, traceEventType, format, state, id);
        }

        // determines if the specified eventType is enabled by the Switch
        private bool IsEnabled(TraceEventType eventType)
        {
            bool enabled = true;

            if (TraceSwitch != null)
            {
                switch (eventType)
                {
                    case TraceEventType.Critical:
                    case TraceEventType.Error:
                        enabled = TraceSwitch.TraceError;
                        break;
                    case TraceEventType.Warning:
                        enabled = TraceSwitch.TraceWarning;
                        break;
                    case TraceEventType.Verbose:
                        enabled = TraceSwitch.TraceVerbose;
                        break;
                    default:
                        enabled = TraceSwitch.TraceInfo;
                        break;
                }
            }

            return enabled;
        }
    }
}
