#region License

/* Copyright (c) 2000-2008 Marc Jacobi
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */

#endregion

using System;
using System.Diagnostics;
using System.Reflection;
using System.Runtime.CompilerServices;

namespace Jacobi.Diagnostics
{
    /// <summary>
    /// The TraceContext class manages method and trace context information.
    /// </summary>
    /// <remarks>The class is intended to be instantiate once in a method that
    /// is to be traced using the 'using' statement. See the example.
    /// </remarks>
    /// <example>
    /// Here is a typical usage scenario:
    /// <code>
    /// public string MyMethod(string param1)
    /// {
    ///     using(TraceContext ctx = new TraceContext())
    ///     {
    ///         ctx.Initialize(this);
    ///         ctx.AddParameterValue(param1, "param1");
    ///         ctx.TraceMethodEntry();
    /// 
    ///         /// ... method body ...
    /// 
    ///         ctx.SetReturnValue(param1);
    ///         return param1;
    /// 
    ///     } // Dispose is called and manages call stack and exit Trace.
    /// }
    /// </code>
    /// </example>
    public class TraceContext : MethodContext
    {
        /// <summary>
        /// Initializes a new instance.
        /// </summary>
        /// <remarks>This instance is pushed as a logical operation
        /// onto the <see cref="Trace.CorrelationManager"/>.
        /// Refer to <see cref="P:TraceManager.LastTraceContext"/> for
        /// more information.</remarks>
        [MethodImpl(MethodImplOptions.NoInlining)]
        public TraceContext()
            : base(1)
        {
            Trace.CorrelationManager.StartLogicalOperation(this);
        }

        /// <summary>
        /// Constructs a base using a stack frame count.
        /// </summary>
        /// <param name="skipFrames">Must be greater than zero.</param>
        protected TraceContext(int skipFrames)
            : base(skipFrames)
        {
            Trace.CorrelationManager.StartLogicalOperation(this);
        }

        /// <summary>
        /// Constructs a base using a known method.
        /// </summary>
        /// <param name="methodBase">Must not be null.</param>
        protected TraceContext(MethodBase methodBase)
            : base(methodBase)
        {
            Trace.CorrelationManager.StartLogicalOperation(this);
        }

        private int _instanceHashCode;
        /// <summary>
        /// Initializes the TraceContext.
        /// </summary>
        /// <param name="type">The type of the calling class.</param>
        /// <param name="instance">The instance of the calling method. 
        /// Can be null for static methods.</param>
        /// <remarks>The method must always be called right after the 
        /// constructor.</remarks>
        public void Initialize(object instance)
        {
            ThrowIfDisposed();

            if (instance != null)
            {
                _instanceHashCode = instance.GetHashCode();
            }

            base.ClassInstance = instance;
            _tracer = TraceManager.Current.GetMethodTracer(instance, MethodBase);
        }

        /// <summary>
        /// Gets the context of the calling 'method'.
        /// </summary>
        /// <remarks>The retruned context is the first method in the call stack that maintains
        /// a <see cref="MethodContext"/> or derived class. It can be null.</remarks>
        public new TraceContext CallingContext
        {
            get { return base.CallingContext as TraceContext; }
        }

        private MethodTracer _tracer;
        /// <summary>
        /// Gets the current <see cref="MethodTracer"/> for this context.
        /// </summary>
        protected internal MethodTracer MethodTracer
        {
            get { return _tracer; }
        }

        /// <summary>
        /// Outputs the method entry trace.
        /// </summary>
        public void TraceMethodEntry()
        {
            ThrowIfDisposed();

            if (MethodTracer.IsTraceEnabled(MethodTraceFlow.Entry))
            {
                MethodTracer.FormatTrace(this, TraceEventType.Start, null,
                    MethodBase, _instanceHashCode);
            }
        }

        /// <summary>
        /// Outputs the method exit trace.
        /// </summary>
        /// <remarks>Called automatically when the <see cref="Dispose"/>
        /// method executes.</remarks>
        public void TraceMethodExit()
        {
            ThrowIfDisposed();

            if (MethodTracer.IsTraceEnabled(MethodTraceFlow.Exit))
            {
                MethodTracer.FormatTrace(this, TraceEventType.Stop, null,
                    MethodBase, _instanceHashCode);
            }
        }

        /// <summary>
        /// Closes down the context.
        /// </summary>
        /// <param name="disposing">Always true.</param>
        /// <remarks>Outputs the method exit trace, flushes the 
        /// <see cref="TraceSource"/> and removes the this intance
        /// from the <see cref="Trace.CorrelationManager"/>.</remarks>
        protected override void Dispose(bool disposing)
        {
            try
            {
                if (disposing)
                {
                    if (MethodTracer != null)
                    {
                        TraceMethodExit();

                        MethodTracer.TraceSource.Flush();

                        _tracer = null;
                    }

                    Trace.CorrelationManager.StopLogicalOperation();
                }
            }
            finally
            {
                base.Dispose(disposing);
            }
        }

        /// <summary>
        /// Returns the current TraceContext or null.
        /// </summary>
        /// <param name="skipFrames">The number of call stack frames to skip
        /// to determine if the TraceContext on the stack belongs to
        /// the calling method.</param>
        /// <returns>Returns null if the TraceContext on the stack does
        /// not belong to the calling method or the stack is empty.</returns>
        [MethodImpl(MethodImplOptions.NoInlining)]
        protected internal static new TraceContext GetCurrent(int skipFrames)
        {
            return MethodContext.GetCurrent(skipFrames + 1) as TraceContext;
        }
    }
}
