#region License

/* Copyright (c) 2000-2008 Marc Jacobi
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */

#endregion

using System;
using System.Collections.Generic;
using System.Reflection;

using Jacobi.Diagnostics.Formatters;

namespace Jacobi.Diagnostics
{
    /// <summary>
    /// Manages a collection of <see cref="ITraceFormatter"/>
    /// implementation.
    /// </summary>
    internal class TraceFormatProvider : ITraceFormatProvider
    {
        private Dictionary<string, ITraceFormatter> _formatters =
            new Dictionary<string, ITraceFormatter>();

        /// <summary>
        /// Initializes the provider with the available <see cref="ITraceFormatter"/>
        /// implementations.
        /// </summary>
        public TraceFormatProvider()
        {
            // TODO: get this from config

            ITraceFormatter formatter = new MethodBaseTraceFormatter();
            formatter.Initialize(this);
            _formatters.Add(typeof(MethodBase).Name, formatter);

            formatter = new ExceptionTraceFormatter();
            formatter.Initialize(this);
            _formatters.Add(typeof(Exception).Name, formatter);

            formatter = new TraceContextStackTraceFormatter();
            formatter.Initialize(this);
            _formatters.Add(typeof(TraceContext).Name, formatter);

            formatter = new ObjectTraceFormatter();
            formatter.Initialize(this);
            _formatters.Add(typeof(Object).Name, formatter);
        }

        private Type GetSupportedBaseType(Type formatType)
        {
            while (formatType != null &&
                !_formatters.ContainsKey(formatType.Name))
            {
                formatType = formatType.BaseType;
            }

            return formatType;
        }

        #region ITraceFormatProvider Members

        /// <summary>
        /// Returns the formatter for the specified <paramref name="formatType"/>
        /// or one of its base classes.
        /// </summary>
        /// <param name="formatType">A type to retrieve the formatter for.</param>
        /// <returns>Returns null when no suitable formatter was found.</returns>
        public ITraceFormatter GetFormatter(Type formatType)
        {
            formatType = GetSupportedBaseType(formatType);

            if (formatType != null)
            {
                return _formatters[formatType.Name];
            }

            return null;
        }

        #endregion
    }
}
