#region License

/* Copyright (c) 2000-2008 Marc Jacobi
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */

#endregion

using System;
using System.Collections.Generic;
using System.Configuration;
using System.Diagnostics;
using System.Reflection;

namespace Jacobi.Diagnostics
{
    /// <summary>
    /// The TraceManager class is a singleton and manages <see cref="TraceSource"/>
    /// and <see cref="MethodTracer"/>instances.
    /// </summary>
    public class TraceManager
    {
        private object _lock = new object();
        private static readonly TraceManager _instance = new TraceManager();
        private TraceManager() {}

        // all trace sources
        private Dictionary<string, TraceSource> _traceSources = 
            new Dictionary<string, TraceSource>();

        // all method tracers
        private Dictionary<string, MethodTracer> _tracers =
            new Dictionary<string, MethodTracer>();

        // the format providers that gives out trace formatters.
        private ITraceFormatProvider _formatProvider = new TraceFormatProvider();

        /// <summary>
        /// Gets the current TraceManager instance.
        /// </summary>
        public static TraceManager Current
        {
            get { return _instance; }
        }

        /// <summary>
        /// Clears all state from the TraceManager.
        /// </summary>
        /// <remarks>This will result in accessing the config file
        /// when new requests for <see cref="MethodTracer"/>s come in.</remarks>
        public void Clear()
        {
            lock (_lock)
            {
                _tracers.Clear();
                _traceSources.Clear();

                _defaultTracer = null;
                _defaultTraceSource = null;
            }
        }

        /// <summary>
        /// Returns the <see cref="TraceSource"/> for the
        /// specified <paramref name="instance"/>.
        /// </summary>
        /// <param name="type">The type of <paramref name="instance"/>.</param>
        /// <param name="instance">An instance or null.</param>
        /// <returns>Never returns null.</returns>
        /// <remarks>The <see cref="TraceSource"/> returned is named
        /// after the full name of the type <typeparamref name="T"/>.
        /// "MyCompany.MyProduct.MyClass"
        /// </remarks>
        public TraceSource GetTraceSource(Type type, object instance)
        {
            Throw.IfArgumentNull(type, "type");

            return GetTraceSource(ReflectionHelper.GetFullName(type));
        }

        private TraceSource GetTraceSource(string sourceName)
        {
            Throw.IfArgumentNullOrEmpty(sourceName, "sourceName");

            TraceSource traceSrc = null;

            if (_traceSources.ContainsKey(sourceName))
            {
                lock (_lock)
                {
                    if (_traceSources.ContainsKey(sourceName))
                    {
                        traceSrc = _traceSources[sourceName];
                    }
                }
            }

            if (!_traceSources.ContainsKey(sourceName))
            {
                traceSrc = new TraceSource(sourceName);

                // test for an unconfigured trace source
                if (traceSrc.Listeners["Default"] != null &&
                    traceSrc.Listeners.Count == 1)
                {
                    traceSrc = NullTraceSource;
                }

                lock (_lock)
                {
                    if (!_traceSources.ContainsKey(sourceName))
                    {
                        _traceSources[sourceName] = traceSrc;
                    }
                    else
                    {
                        traceSrc = _traceSources[sourceName];
                    }
                }
            }

            return traceSrc;
        }

        /// <summary>
        /// Returns the <see cref="MethodTracer"/> for the specified
        /// <paramref name="methodName"/>.
        /// </summary>
        /// <param name="type">The type of the calling class.</param>
        /// <param name="instance">Instance of the calling class. Can be null.</param>
        /// <param name="memberInfo">The method info of the calling client method.</param>
        /// <returns>Never returns null.</returns>
        public MethodTracer GetMethodTracer(object instance, MemberInfo memberInfo)
        {
            Throw.IfArgumentNull(memberInfo, "memberInfo");

            MethodTracer tracer = null;
            string methodFullName = ReflectionHelper.GetFullName(memberInfo);
            Type type = memberInfo.DeclaringType;

            if (_tracers.ContainsKey(methodFullName))
            {
                lock (_lock)
                {
                    if (_tracers.ContainsKey(methodFullName))
                    {
                        tracer = _tracers[methodFullName];
                    }
                }
            }

            if (!_tracers.ContainsKey(methodFullName))
            {
                // retrieve the trace source for the type
                TraceSource traceSrc = GetTraceSource(type, instance);

                if (traceSrc != NullTraceSource)
                {
                    // create a TraceSwitch for this method
                    TraceSwitch traceSwitch =
                        CreateMethodTraceSwitch(memberInfo, TraceLevel.Verbose);

                    // create the method tracer
                    tracer = new MethodTracer(traceSrc, traceSwitch);
                }
                else
                {
                    // create the method tracer
                    tracer = new MethodTracer(traceSrc, null);
                }

                lock(_lock)
                {
                    if (!_tracers.ContainsKey(methodFullName))
                    {
                        _tracers[methodFullName] = tracer;
                    }
                    else
                    {
                        tracer = _tracers[methodFullName];
                    }
                }
            }

            return tracer;
        }

        // creates a new TraceSwitch for the specified method
        private TraceSwitch CreateMethodTraceSwitch(
            MemberInfo memberInfo, TraceLevel defaultTraceLevel)
        {
            Throw.IfArgumentNull(memberInfo, "memberInfo");

            TraceSwitch traceSwitch = null;
            SwitchAttribute switchAttr = (SwitchAttribute)
                Attribute.GetCustomAttribute(memberInfo, typeof(SwitchAttribute));

            // no SwitchAttribute declared on method 
            if (switchAttr == null)
            {
                string methodFullName = ReflectionHelper.GetFullName(memberInfo);
                
                traceSwitch = new MethodTraceSwitch(
                        methodFullName, methodFullName, defaultTraceLevel.ToString());
            }
            else
            {
                string switchName = switchAttr.SwitchName;
                string description = switchAttr.SwitchDescription;

                if (String.IsNullOrEmpty(description))
                {
                    description = ReflectionHelper.GetFullName(memberInfo);
                }

                traceSwitch = (TraceSwitch)Activator.CreateInstance(
                    switchAttr.SwitchType, new object[] { 
                        switchName, description, defaultTraceLevel.ToString() });
            }

            // trigger the switch initialization from config.
            defaultTraceLevel = traceSwitch.Level;

            return traceSwitch;
        }

        private MethodTracer _defaultTracer;
        /// <summary>
        /// Gets the default <see cref="MethodTracer"/> instance.
        /// </summary>
        /// <remarks>
        /// The default <see cref="MethodTracer"/> is used when one of the
        /// static <see cref="MethodTracer"/> methods is called without a
        /// <see cref="TraceContext"/> being present.
        /// </remarks>
        public MethodTracer DefaultMethodTracer
        {
            get
            {
                if (_defaultTracer == null)
                {
                    lock (_lock)
                    {
                        if (_defaultTracer == null)
                        {
                            _defaultTracer = new MethodTracer(DefaultTraceSource, null);
                        }
                    }
                }

                return _defaultTracer;
            }
        }

        private TraceSource _defaultTraceSource;
        /// <summary>
        /// Returns the default <see cref="TraceSource"/> instance.
        /// </summary>
        /// <remarks>
        /// The <see cref="TraceSource"/> is named after the
        /// full name of the entry assembly (the application).
        /// </remarks>
        public TraceSource DefaultTraceSource
        {
            get
            {
                if (_defaultTraceSource == null)
                {
                    lock (_lock)
                    {
                        if (_defaultTraceSource == null)
                        {
                            Assembly execAssem = Assembly.GetEntryAssembly();

                            _defaultTraceSource = GetTraceSource(
                                execAssem.GetName().Name);
                        }
                    }
                }

                return _defaultTraceSource;
            }
        }

        /// <summary>
        /// Returns the most recent <see cref="TraceContext"/> instance
        /// from the <see cref="Trace.CorrelationManager"/> stack.
        /// </summary>
        public TraceContext LastTraceContext
        {
            get
            {
                TraceContext ctx = null;

                foreach (object item in 
                    Trace.CorrelationManager.LogicalOperationStack.ToArray())
                {
                    ctx = item as TraceContext;

                    if (ctx != null)
                    {
                        break;
                    }
                }

                return ctx;
            }
        }

        private TraceSource _nullTraceSource;
        /// <summary>
        /// Gets a <see cref="TraceSource"/> that has no listseners.
        /// </summary>
        public TraceSource NullTraceSource
        {
            get
            {
                if (_nullTraceSource == null)
                {
                    lock (_lock)
                    {
                        if (_nullTraceSource == null)
                        {
                            _nullTraceSource = new TraceSource(
                                "NullTraceSource", SourceLevels.Off);
                            _nullTraceSource.Listeners.Clear();
                        }
                    }
                }

                return _nullTraceSource;
            }
        }

        /// <summary>
        /// Formats the specified <paramref name="state"/> object
        /// into human readable text.
        /// </summary>
        /// <param name="context">The TraceContext. Can be null.</param>
        /// <param name="state">A state object. Can be null.</param>
        /// <returns>Returns a trace text string.</returns>
        /// <remarks>The method finds an <see cref="ITraceFormatter"/> implementation
        /// to format the <paramref name="state"/> object. If the 
        /// <paramref name="state"/> object is null, an empty string is returned.</remarks>
        public string Format(TraceContext context, object state)
        {
            if (state != null)
            {
                ITraceFormatter formatter = 
                    _formatProvider.GetFormatter(state.GetType());

                if (formatter != null)
                {
                    return formatter.Format(context, state);
                }
            }

            return String.Empty;
        }
    }
}
