#region License

/* Copyright (c) 2000-2008 Marc Jacobi
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */

#endregion

using System;
using System.Globalization;

namespace Jacobi
{
    /// <summary>
    /// Implements the Dispose implementation pattern as a base class.
    /// </summary>
    public abstract class DisposableBase : IDisposable
    {
        #region IDisposable Members
        /// <summary>
        /// Call to dispose of this instance.
        /// </summary>
        public void Dispose()
        {
            Dispose(true);
        }

        /// <summary>
        /// Called either from <see cref="Dispose"/> or the Finalizer (not in this base class)
        /// to dispose of this instance.
        /// </summary>
        /// <param name="disposing">True when called from <see cref="Dispose"/>. 
        /// False when called from the Finalizer.</param>
        protected virtual void Dispose(bool disposing)
        {
            //if (!_disposed)
            //{
            //    if (disposing)
            //    {
            //        // dispose managed resources
            //    }

            //    // dispose unmanaged resources
            //}

            _disposed = true;
        }

        private bool _disposed;
        /// <summary>
        /// Gets a value indicating if this instance has been disposed.
        /// </summary>
        protected bool IsDisposed
        {
            get { return _disposed; }
        }

        /// <summary>
        /// Throws an exception if the instance has been disposed.
        /// </summary>
        /// <exception cref="ObjectDisposedException">Throw when the instance is disposed.</exception>
        protected virtual void ThrowIfDisposed()
        {
            if (IsDisposed)
            {
                throw new ObjectDisposedException(String.Format(
                    CultureInfo.InvariantCulture,
                    "This instance of type {0} has been disposed.",
                    GetType().Name));
            }
        }
        #endregion
    }
}
