#region License

/* Copyright (c) 2000-2008 Marc Jacobi
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a copy 
 * of this software and associated documentation files (the "Software"), to 
 * deal in the Software without restriction, including without limitation the 
 * rights to use, copy, modify, merge, publish, distribute, sublicense, and/or 
 * sell copies of the Software, and to permit persons to whom the Software is 
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included in 
 * all copies or substantial portions of the Software. 
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, 
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE 
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER 
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN 
 * THE SOFTWARE.
 */

#endregion

using System;
using System.Collections.Generic;
using System.Globalization;
using System.Xml;

namespace Jacobi
{
    /// <summary>
    /// The Throw class provides static helper methods for method parameter validation.
    /// </summary>
    public static class Throw
    {
        /// <summary>
        /// Checks if the parameter <paramref name="value"/> is null.
        /// </summary>
        /// <param name="value">The parameter value.</param>
        /// <param name="paramName">The parameter name.</param>
        /// <exception cref="ArgumentNullException">Thrown when <paramref name="value"/> is null.</exception>
        public static void IfArgumentNull(object value, string paramName)
        {
            if (value == null)
            {
                throw new ArgumentNullException(paramName);
            }
        }

        /// <summary>
        /// Checks if the parameter <paramref name="value"/> is null.
        /// </summary>
        /// <param name="value">The parameter value.</param>
        /// <param name="paramName">The parameter name.</param>
        /// <param name="message">An exception message.</param>
        /// <exception cref="ArgumentNullException">Thrown when <paramref name="value"/> is null.</exception>
        public static void IfArgumentNull(object value, string paramName, string message)
        {
            if (value == null)
            {
                throw new ArgumentNullException(paramName, message);
            }
        }

        /// <summary>
        /// Checks if the parameter <paramref name="value"/> is empty.
        /// </summary>
        /// <param name="value">The parameter value.</param>
        /// <param name="paramName">The parameter name.</param>
        /// <exception cref="ArgumentException">Thrown when <paramref name="value"/> is empty.</exception>
        public static void IfArgumentEmpty(Guid value, string paramName)
        {
            if (value == Guid.Empty)
            {
                throw new ArgumentException(Properties.Resources.Check_ArgumentEmpty, paramName);
            }
        }

        /// <summary>
        /// Checks if the parameter <paramref name="value"/> is null or empty.
        /// </summary>
        /// <param name="value">The parameter value.</param>
        /// <param name="paramName">The parameter name.</param>
        /// <exception cref="ArgumentNullException">Thrown when <paramref name="value"/> is null.</exception>
        /// <exception cref="ArgumentException">Thrown when <paramref name="value"/> is empty.</exception>
        public static void IfArgumentNullOrEmpty(string value, string paramName)
        {
            if (String.IsNullOrEmpty(value))
            {
                IfArgumentNull(value, paramName);

                throw new ArgumentException(Properties.Resources.Check_ArgumentEmpty, paramName);
            }
        }

        /// <summary>
        /// Checks if the parameter <paramref name="value"/> is null or empty.
        /// </summary>
        /// <param name="value">The parameter value.</param>
        /// <param name="paramName">The parameter name.</param>
        /// <exception cref="ArgumentNullException">Thrown when <paramref name="value"/> is null.</exception>
        /// <exception cref="ArgumentException">Thrown when <paramref name="value"/> is empty.</exception>
        public static void IfArgumentNullOrEmpty(XmlQualifiedName value, string paramName)
        {
            IfArgumentNull(value, paramName);

            if (value.IsEmpty)
            {
                throw new ArgumentException(Properties.Resources.Check_ArgumentEmpty, paramName);
            }
        }

        /// <summary>
        /// Checks if the parameter <paramref name="value"/> is out of range.
        /// </summary>
        /// <typeparam name="T">The parameter data type.</typeparam>
        /// <param name="value">The parameter value.</param>
        /// <param name="minValue">The parameter's minimal value.</param>
        /// <param name="maxValue">The parameter's maximal value.</param>
        /// <param name="paramName">The parameter name.</param>
        /// <exception cref="ArgumentOutOfRangeException">Thrown when <paramref name="value"/> is out of range.</exception>
        public static void IfArgumentOutOfRange<T>(IComparable<T> value, T minValue, T maxValue, string paramName)
        {
            if (Check.IsOutOfRange<T>(value, minValue, maxValue))
            {
                throw new ArgumentOutOfRangeException(paramName, value, 
                    String.Format(CultureInfo.InvariantCulture,
                        Properties.Resources.Check_ArgumentOutOfRange, minValue, maxValue));
            }
        }

        /// <summary>
        /// Checks if the parameter <paramref name="value"/> is a certain <see cref="Type"/>.
        /// </summary>
        /// <typeparam name="T">The <see cref="Type"/> of the parameter.</typeparam>
        /// <param name="value">The parameter value.</param>
        /// <param name="paramName">The parameter name.</param>
        /// <exception cref="ArgumentException">Thrown when <paramref name="value"/> is not of Type <b>T</b>.</exception>
        public static void IfArgumentNotOfType<T>(object value, string paramName)
		{
			if (!(value is T))
			{
				throw new ArgumentException(
					String.Format(CultureInfo.InvariantCulture,
                        Properties.Resources.Check_ArgumentNotOfType, typeof(T).FullName), paramName);
			}
		}

        /// <summary>
        /// Throws an exception if the parameter <paramref name="value"/> is not null (set).
        /// </summary>
        /// <param name="value">Th parameter value.</param>
        /// <param name="paramName">The parameter name.</param>
        /// <exception cref="InvalidObjectException">Thrown when the parameter <paramref name="value"/> 
        /// is set to in instance of an object.</exception>
        public static void InvalidOperationIfSet(object value, string paramName)
        {
            if (value != null)
            {
                throw new InvalidOperationException(
                    String.Format(CultureInfo.InvariantCulture,
                        Properties.Resources.Check_InvalidOperationIfSet, paramName));
            }
        }
	}
}
